/*************************************************
    DDSPLLU.C:
               Sample driver for the YAPLL-U board
               with an optional YADDS-1 used as
               a frequency reference
    Tibor Bece
 *************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <conio.h>

/* printer port addresses: */
#define LPT1 0x378
#define LPT2 0x278
#define LPT3 0x3bc

/*
  D25 pinout for the YADDS series boards:
     pin2-Data
     pin3-Clk
     pin4-YADDS Shift Enable
     pin5-YADDS Transfer pulse
     pin6-ADC Enable
     pin7-PLL Enable
     pin8-Aux PLL Enable
     pin18-Gnd
*/

#define SD 0x01
#define SCLK 0x02
#define SH_EN 0x04
#define TXFR 0x08
#define PLL_EN 0x20

/* MC145191 IC related things: */
#define C_BITS 8
#define R_BITS 16
#define N_BITS 24
  /* C register bits: */
#define PD_TRUE 0x00L
#define PD_INVERT 0x80L
#define SEL_PA 0x40L
#define SEL_PB 0x00L
#define LD_EN 0x20L
  /* programmable charge pump current: */
#define I100 0x0CL
#define I90  0x08L
#define I80  0x04L
#define I70  0x00L
#define I75  0x0AL
#define I50  0x06L
#define I25  0x02L
  /* limits on main counter divisor (before split up to 'A' and 'N': */
#define MIN_N 4032L
#define MAX_N 262143L

  /* additional bits in 'R' register: */
#define CRYSTAL_MODE 0x02000L  /* enable input amplifier even for 'x_ext'*/
  /* additional bits in 'A' register: */
#define OUTA_PORT 0x300000L
#define OUTA_DOUT 0x700000L
#define OUTA_FV   0xB00000L
#define OUTA_FR   0xF00000L

/* printer port related global data: */
int lpt_base=LPT2;  /* hold printer port address here */
int lpt_dat=0;      /* save lpt data byte here  */

/* global data related to the PLL board: */
double freq;    /* in MHz */
double offset;  /* in MHz */
double hop;     /* in MHz */
double PLL_xtal; /* MHz */
double PLL_fref; /* MHz */
long PLL_R;
long PLL_C;
long oldPLL_N=0;
double DDS_xref; /* MHz */

/****** Low level printer port driving routines:******/
void enb_PLL (int d)
/* set/reset YAPLL enable line (prepare for data transfer): */
{
  lpt_dat=inportb(lpt_base)&~(SD | SCLK | PLL_EN);  /* prepare data.. */
  outportb (lpt_base,((d)?(lpt_dat|PLL_EN):(lpt_dat)));     /* write. */
}

void clk_bit (int d)
/* clock in one bit to YAPLL: */
{
  /* setup data: */
  outportb (lpt_base,((d)?(lpt_dat|SD|PLL_EN):(lpt_dat|PLL_EN)));
  /* add a positive clock: */
  outportb (lpt_base,((d)?
           (lpt_dat|SD|SCLK|PLL_EN):(lpt_dat|SCLK|PLL_EN)));
}

void load_PLL (long data, int bit_no)
/* load 'bits_no' bits of 'data', MSB first: */
{
  int i;
  long bitmask;
  bitmask = 1L << (bit_no-1);     /* prepare mask for first bit */
  enb_PLL (1);                    /* set Enable line of YAPLL... */
  for (i=0; i<bit_no; i++) {
    clk_bit ((data & bitmask)?1:0);    /* load current bit,... */
    bitmask >>= 1;                     /* set mask for next bit */
  }
  enb_PLL (0);                    /* reset Enable line. */
}

/*** routines to program the 'C', 'R' and 'N' registers: ***/
void load_C (long data)
{
  load_PLL (data, C_BITS);
}

void load_R (long data)
{
  data &= 0x1FFFL;        /* leave 13 bits only */
  data |= CRYSTAL_MODE;   /* add PLL chip 'mode' bits to 'R' */
  load_PLL (data, R_BITS);
}

void load_N (long data)
{
  long PLL_A_reg, PLL_N_reg;

    /* split main divider value to dual modulus components: */
  PLL_A_reg = data & 0x00003FL;
  PLL_N_reg = (data << 2) & 0x0FFF00L;
    /* collect all the necessary bits: */
  data = OUTA_FV | PLL_N_reg | PLL_A_reg;
  load_PLL (data, N_BITS);
}

/****** loading the YADDS board: ******/

long f2N (double f)
/* convert frequency to a DDS 32 bit phase increment word: */
{
  return ((long)(f/DDS_xref*4294967296.0) );
}

void txfr_DDS()
/* generate transfer pulse for YADDS-1: */
{
  int w;
  w=inportb (lpt_base);
  outportb (lpt_base, w|TXFR); /* generate transfer pulse */
  outportb (lpt_base, w);
}

void load_DDS_N (long N)
/* N : 32 bit DDS frequency control word */
/* -the transfer pulse is generated separately, in order
    to minimise the delay between DDS load and PLL load */
{
  int i,d,w;

  w = inportb (lpt_base) & ~(SD | SCLK | TXFR);
  outportb (lpt_base, w);
  /* load 32 bits of DDS frequency word, LSB first: */
  for (i=0; i<32; i++) {
    d = N & 0x00000001;
    if (d) {
      outportb (lpt_base, w|SH_EN|SD);
      outportb (lpt_base, w|SH_EN|SD|SCLK);
    }
    else {
      outportb (lpt_base, w|SH_EN);
      outportb (lpt_base, w|SH_EN|SCLK);
    }
    N = N >> 1;
  }
  outportb (lpt_base, w);
}

/***** Main frequency setting routine *****/

int set_freq (double MHz)
{
  long PLL_N;
  double DDS_freq;
  int i;
  int fail=0;

  PLL_N = (long) (MHz/PLL_fref+0.5);        /* round to nearest integer */
  if ((PLL_N >= MIN_N)&&(PLL_N <= MAX_N)) { /* check valid N range */
    DDS_freq = MHz / PLL_N * PLL_R;         /* required exact 'xtal' freq. */
    if (PLL_N == oldPLL_N) {
      /* do not reload PLL with same value */
      load_DDS_N (f2N(DDS_freq));
      txfr_DDS();
    }
    else {
      oldPLL_N = PLL_N;
      load_DDS_N (f2N(DDS_freq));
      load_C (PLL_C);
      load_R (PLL_R);
      load_N (PLL_N); /* minimise delay between PLL and DDS load: */
      txfr_DDS();
    }
  }
  else {
    printf ("Warning: PLL 'N' out of range! (N=%ld)\n",PLL_N);
    fail=1;
  }
  return (fail);
}

/****** User interface: ******/

void help ()
{
  puts ("Available commads:");
  puts ("  [f]<new frequency,MHz> ('f' is optional)");
  puts ("  o<new frequency offset,MHz>");
  puts ("  r<new PLL 'R' value>   (set PLL reference divisor)");
  puts ("  c<new PLL 'C' value>   (set PLL config byte in hex)");
  puts ("  x<new PLL xtal frequency> (also nominal DDS output!)");
  puts ("  i<%charge pump current>  (valid range: 25..100%)");
  puts ("  h<hop MHz>             (check PLL transient response)");
  puts ("");
}

int get_command (char *s)
/* see what to do with the string just typed in... */
{
  int quit=0;
  int PD_current;
  int hop_delay;

  if (s[0]=='f') {       /* read frequency */
    sscanf (&s[1],"%lf",&freq);
    printf ("new frequency: %lfMHz\n",freq);
  }
  else if (s[0]=='o') {  /* read frequency offset */
    sscanf (&s[1],"%lf",&offset);
    printf ("new frequency offset: %lfMHz\n",offset);
  }
  else if (s[0]=='r') {  /* set new PLL 'R' reference divisor: */
    if (1==sscanf (&s[1],"%ld",&PLL_R)) {
      if (PLL_R < 1)
        PLL_R = 1;
      PLL_fref=PLL_xtal/PLL_R;
      oldPLL_N = 0; /* force reloading the PLL */
      printf ("new PLL reference frequency: %lfMHz\n",PLL_fref);
    }
  }
  else if (s[0]=='c') {  /* set new PLL 'C' control bits: */
    sscanf (&s[1],"%lx",&PLL_C);
    printf ("new PLL control byte: 0x%02lx\n",PLL_C);
    oldPLL_N = 0; /* force reloading the PLL */
  }
  else if (s[0]=='x') {  /* set new PLL xtal frequency: */
    sscanf (&s[1],"%lf",&PLL_xtal);
    oldPLL_N = 0; /* force reloading the PLL */
    PLL_fref = PLL_xtal/PLL_R;
    printf ("new PLL crystal frequency: %lfMHz\n",PLL_xtal);
  }
  else if (s[0]=='i') {  /* set charge pump current: */
    if (1 == sscanf (&s[1],"%d",&PD_current)) {
      PLL_C &= 0x0F1L; /* clear current setting bits */
      if (PD_current <= 35) {
        PD_current = 25;
        PLL_C |= I25;
      }
      else if (PD_current <= 60) {
        PD_current = 50;
        PLL_C |= I50;
      }
      else if (PD_current <= 72) {
        PD_current = 70;
        PLL_C |= I70;
      }
      else if (PD_current <= 78) {
        PD_current = 75;
        PLL_C |= I75;
      }
      else if (PD_current <= 84) {
        PD_current = 80;
        PLL_C |= I80;
      }
      else if (PD_current <= 94) {
        PD_current = 90;
        PLL_C |= I90;
      }
      else {
        PD_current = 100;
        PLL_C |= I100;
      }
      oldPLL_N = 0; /* force reloading the PLL */
      printf ("new charge pump current: %d\%\n",PD_current);
    }
  }
  else if (s[0]=='q') {
    quit=1;
  }
  else if (s[0]=='?') {
    help();
  }
  /* frequency hopping, for checking PLL transient response: */
  else if (s[0]=='h') {
    hop_delay=PLL_R/PLL_xtal/2;
    sscanf (&s[1],"%lf",&hop);
    printf ("frequency hop %lfMHz <-> %lfMHz...\n",freq, freq+hop);
    while (!kbhit()) {
      set_freq(freq+offset);
      delay(hop_delay);
      set_freq(freq+offset+hop);
      delay(hop_delay);
    }
    set_freq(freq+offset);
  }
  else { /* interpret just a number as frequency data */
    if (1==sscanf (s,"%lf",&freq)) {
      printf ("new frequency: %lfMHz\n",freq);
    }
  }
  return (quit);
}

#define PROMPT "*>"
#define TITLE "\n** DDS/PLLU driver: YADDS-1 driven YAPLL-U **\n"

int main ()
{
  /*
    Simple main program - driving the YAPLL-U board,
    by default the external reference is supplied by YADDS-1
    at arouond 10.7MHz:
  */

  int quit=0;
  char s[256];  /* hold user input here */

    /* initial values: */
  freq=650.0;   /* default output frequency */
  PLL_xtal = 10.7;
  PLL_R = 107;
  PLL_fref = PLL_xtal / PLL_R;
  PLL_C = PD_TRUE | SEL_PA | I90;     /* set bits in C register */
    /* note: YAPLL-U board is configured as follows:
       -POS-1025 VCO module used
       -target reference frequency: 100kHz, PLL wc: 2kHz
       -phase comparator A used (default on PCB)
       -additional active lowpass filter used gf=20kHz (inverting!)
       -loop gain set to require 100% charge pump current at 1GHz
    */
  DDS_xref=50.000;    /* crystal frequency of the YADDS board */
  offset=0.0;
  hop=0.0;

  puts (TITLE);
  while (!quit) {                   /* loop until 'quit':  */
    printf (PROMPT);                    /* print prompt, */
    gets (s);  fflush(stdin);           /* get response, */
    quit = get_command (s);             /* interpret it, */
    set_freq(freq+offset);              /* drive the YAPLL */
  }
  return (0);
}